------------------
-- MiniMap class
__MiniMap            = ZO_Object:Subclass()
__MiniMap.savedVars  = nil
__MiniMap.data       = {}
__MiniMap.zoom       = 0.5
__MiniMap.resize     = false
__MiniMap.fastTravel = false

function __MiniMap:New( ... )
	local result = ZO_Object.New( self )
	result:Initialize( ... )
	return result
end

function __MiniMap:Initialize( control )
	self.control = control

	-- load settings
	local defaults = {
		offsetX         = self.control:GetRight() - GuiRoot:GetRight(),
		offsetY         = self.control:GetBottom() - GuiRoot:GetBottom(),
		width           = 260,
		height          = 260,
		defaultZoom     = 0.5,
		defaultPinScale = 1,
	}
	self.savedVars = ZO_SavedVars:NewAccountWide( string.format( '%s_%s', _LUI.addOnName, 'SavedVariables' ), _LUI.savedVarsID, 'MiniMap', defaults )
	self.zoom = self.savedVars.defaultZoom
	self.control:ClearAnchors()
	self.control:SetAnchor( BOTTOMRIGHT, GuiRoot, BOTTOMRIGHT, self.savedVars.offsetX, self.savedVars.offsetY )
	self.control:SetDimensions( self.savedVars.width, self.savedVars.height )
	self.control:GetNamedChild( '_Background' ):SetDimensions( self.savedVars.width + 8, self.savedVars.height + 8 )
	self.control:GetNamedChild( '_Scroll' ):SetDimensions( self.savedVars.width, self.savedVars.height )
	self.control:SetHandler( 'OnMoveStop', function()
		self.savedVars.offsetX = self.control:GetRight() - GuiRoot:GetRight()
		self.savedVars.offsetY = self.control:GetBottom() - GuiRoot:GetBottom()
	end )
	self.control:SetHandler( 'OnResizeStart', function() self.resize = true  end )
	self.control:SetHandler( 'OnResizeStop',  function() self.resize = false end )

	-- set control defaults
	MiniMap_Scroll:SetScrollBounding( 0 )
	MiniMap_PlayerCam:SetDimensions( 96, 96 )
	MiniMap_PlayerCam:SetAddressMode( TEX_MODE_CLAMP )
	MiniMap_PlayerCam:SetBlendMode( TEX_BLEND_MODE_ADD )
	MiniMap_PlayerCam:SetColor( _LUI.config.colors.WHITE.r, _LUI.config.colors.WHITE.g, _LUI.config.colors.WHITE.b, MiniMap_PlayerCam:GetAlpha() )
	MiniMap_ZoomLabel:SetText( string.format( '%.0f%%', self.zoom * 100 ) )

	-- set event handlers
	zo_callLater( function() self:OnWorldMapUpdate( 'Initialize', 0 ) end, 1000 )
	self.control:SetHandler( 'OnMouseWheel', function( ... ) self:OnMouseScroll( ... ) end )
	self.control:SetHandler( 'OnMouseDown',  function( ... ) self:OnMouseDown( ... )   end )
	self.control:RegisterForEvent( EVENT_START_FAST_TRAVEL_INTERACTION,      function( nodeIndex ) self:FastTravel() end )
	self.control:RegisterForEvent( EVENT_START_FAST_TRAVEL_KEEP_INTERACTION, function( keepId ) self:FastTravel() end )
	self.control:RegisterForEvent( EVENT_END_FAST_TRAVEL_INTERACTION,        function() self:FastTravel() end )
	self.control:RegisterForEvent( EVENT_END_FAST_TRAVEL_KEEP_INTERACTION,   function() self:FastTravel() end )
end

function __MiniMap:FastTravel()
	self.fastTravel = not self.fastTravel
end

function __MiniMap:OnResize()
	local width  = self.control:GetWidth()
	local height = self.control:GetHeight()
	self.savedVars.width  = ( width  < 100 ) and 100 or width
	self.savedVars.height = ( height < 100 ) and 100 or height
	self.control:SetDimensions( self.savedVars.width, self.savedVars.height )
	self.control:GetNamedChild( '_Background' ):SetDimensions( self.savedVars.width + 8, self.savedVars.height + 8 )
	self.control:GetNamedChild( '_Scroll' ):SetDimensions( self.savedVars.width, self.savedVars.height )
end

function __MiniMap:OnMouseScroll( _, delta, ctrl, alt, shift )
	self:Zoom( delta )
end

function __MiniMap:OnMouseDown( _, button, ctrl, alt, shift )
	if ( button == 3 ) then -- MMB
		self:Zoom( 0 )
	end
end

function __MiniMap:OnUpdate()
	-- limit update speed
	if not _LUI:DelayBuffer( 'MiniMapOnUpdate', 40 ) then return end

	-- check for resize mode
	if ( self.resize ) then
		self:OnResize()
	end

	-- set map to player location
	if not self.fastTravel then
		local setMap = SetMapToPlayerLocation()

		-- check setMap result
		if ( setMap == SET_MAP_RESULT_MAP_CHANGED or setMap == SET_MAP_RESULT_FAILED ) then
			if not self:OnWorldMapUpdate( 'MapChanged', 0 ) then return end
		-- check mapname change
		elseif ( _LUI:DelayBuffer( 'MiniMapMapNameChanged', 5000 ) and self.data.map.rawName ~= GetMapName() ) then
			if not self:OnWorldMapUpdate( 'MapNameChanged', 0 ) then return end
		end
	end

	-- Build map
	if _LUI:DelayBuffer( 'MiniMapBuildMap', 500 ) then
		self:OnBuildMap()
	end

	-- set player position
	local px, py, ph = GetMapPlayerPosition( 'player' )
	local pch        = GetPlayerCameraHeading()
	local hPos       = ( px * ( self.zoom * self.data.map.width  ) ) - ( MiniMap_Scroll:GetWidth() / 2 )
	local vPos       = ( py * ( self.zoom * self.data.map.height ) ) - ( MiniMap_Scroll:GetHeight() / 2 )
	MiniMap_Scroll:SetHorizontalScroll( hPos )
	MiniMap_Scroll:SetVerticalScroll( vPos )
	MiniMap_Player:SetTextureRotation( ph )
	MiniMap_PlayerCam:SetTextureRotation( pch )
end

function __MiniMap:OnWorldMapUpdate( reason, recursion )
	-- reset OnUpdate handler
	self.control:SetHandler( 'OnUpdate', nil )

	-- show status overlay
	MiniMap_Scroll_StatusOverlay:SetHidden( false )
	MiniMap_Scroll_StatusOverlay_Label:SetText( 'Loading' )

	-- count recursion
	recursion = recursion + 1

	-- update world map data
	--CONSOLE:LogInfo( string.format( 'OnWorldMapUpdate called [%d]: %s', recursion, reason ) )
	ZO_WorldMap_UpdateMap()

	-- clear tiles
	local numChildren = MiniMap_Scroll_Map:GetNumChildren()
	for num = 1, numChildren - 1 do
		local tile = MiniMap_Scroll_Map:GetNamedChild( string.format( '_%d', num ) )
		if ( tile ~= nil ) then
			tile:ClearAnchors()
			tile:SetHidden( true )
		end
	end

	-- clear pins
	local numChildren = MiniMap_Scroll_Map_Pins:GetNumChildren()
	for num = 1, numChildren do
		local pin = MiniMap_Scroll_Map_Pins:GetChild( num )
		if ( pin.Zone ~= self.data.map.rawName ) then
			pin:ClearAnchors()
			pin:SetHidden( true )
		end
	end

	-- get map data
	self.data.map             = {}
	self.data.map.textures    = {}
	self.data.map.rawName     = GetMapName()
	self.data.map.tileTexture = GetMapTileTexture()
	self.data.map.dx,
	self.data.map.dy          = GetMapNumTiles()
	self.data.map.numTiles    = self.data.map.dx * self.data.map.dy

	-- set zone name
	MiniMap_Zone:SetText( _LUI:ParseLanguageString( self.data.map.rawName ) )

	-- check numTiles
	if ( self.data.map.numTiles == 0 ) then
		MiniMap_Scroll_StatusOverlay_Label:SetText( string.format( 'Loading map info [%d]', recursion ) )
		if ( recursion < 10 ) then
			zo_callLater( function() self:OnWorldMapUpdate( string.format( 'Map info reload [%d]', recursion ), recursion ) end, 1000 * recursion )
		else
			MiniMap_Scroll_StatusOverlay_Label:SetText( 'Loading failed' )
		end
		return false
	end

	-- store tile textures in table
	if ( self.data.map.numTiles > 1 ) then
		for num = 0, self.data.map.numTiles - 1 do
			local value = string.gsub( self.data.map.tileTexture, '_(%d+).dds', string.format( '_%d.dds', num ) )
			table.insert( self.data.map.textures, value )
		end
	elseif ( self.data.map.numTiles == 1 ) then
		table.insert( self.data.map.textures, self.data.map.tileTexture )
	end

	-- create / reset tile controls
	-- load textures
	for num, texture in pairs( self.data.map.textures ) do
		local tile = nil
		if ( MiniMap_Scroll_Map:GetNamedChild( string.format( '_%d', num ) ) == nil ) then
			tile = WINDOW_MANAGER:CreateControl( string.format( 'MiniMap_Scroll_Map_%s', num ), MiniMap_Scroll_Map, CT_TEXTURE )
		else 
			tile = MiniMap_Scroll_Map:GetNamedChild( string.format( '_%d', num ) )
		end
		tile:SetDimensions( 260, 260 )
		tile:SetTexture( texture )
		self.data.map.tileWidth, self.data.map.tileHeight = tile:GetTextureFileDimensions()
	end

	-- check textures
	local IsTextureLoaded = false
	for num = 1, self.data.map.numTiles do
		local tile = MiniMap_Scroll_Map:GetNamedChild( string.format( '_%d', num ) )
		IsTextureLoaded = tile:IsTextureLoaded()
	end

	if not IsTextureLoaded then
		MiniMap_Scroll_StatusOverlay_Label:SetText( string.format( 'Loading textures [%d]', recursion ) )
		if ( recursion < 10 ) then
			zo_callLater( function() self:OnWorldMapUpdate( string.format( 'Texture reload [%d]', recursion ), recursion ) end, 1000 * recursion )
		else
			MiniMap_Scroll_StatusOverlay_Label:SetText( 'Loading failed' )
		end
		return false
	else
		-- set map dimensions
		self.data.map.width      = self.data.map.tileWidth  * self.data.map.dx
		self.data.map.height     = self.data.map.tileHeight * self.data.map.dy

		-- set update handler
		self.control:SetHandler( 'OnUpdate', function() self:OnUpdate() end )

		-- build map
		self:OnBuildMap()

		-- hide status overlay
		MiniMap_Scroll_StatusOverlay:SetHidden( true )
		return true
	end
end

function __MiniMap:OnBuildMap()
	-- check numTiles
	if ( self.data.map.numTiles == 0 ) then return end

	-- build
	self:BuildMap()
	self:BuildPins()
end

function __MiniMap:BuildMap()
	-- recalc dimensions by zoom
	local tileWidth  = self.zoom * self.data.map.tileWidth
	local tileHeight = self.zoom * self.data.map.tileHeight
	local mapWidth   = self.zoom * self.data.map.width
	local mapHeight  = self.zoom * self.data.map.width

	-- set scroll zoomed dimensions
	MiniMap_Scroll_Map:SetDimensions( mapWidth, mapHeight )

	-- arrange zoomed tiles
	for num = 1, self.data.map.numTiles do
		local tile = MiniMap_Scroll_Map:GetNamedChild( string.format( '_%d', num ) )
		tile:SetHidden( false )
		tile:SetDimensions( tileWidth, tileHeight )
		tile:ClearAnchors()
		tile:SetAnchor( TOPLEFT, MiniMap_Scroll_Map, TOPLEFT, tileWidth * ( ( num - 1 ) % self.data.map.dx ), tileHeight * math.floor( ( num - 1 ) / self.data.map.dx ) )
	end
end

function __MiniMap:BuildPins()
	-- get player
	local pZone = GetUnitZone( 'player' )
	local pName = GetRawUnitName( 'player' )

	-- display pins
	local numChildren = ZO_WorldMapContainer:GetNumChildren()
	for num = 1, numChildren do
		local wpin = ZO_WorldMapContainer:GetChild( num )
		if ( wpin.m_Pin ~= nil and wpin.m_Pin:GetUnitTag() ~= 'player' and wpin.m_Pin.normalizedX ~= nil ) then
			-- get pin
			local pinName = wpin:GetName()
			local pin     = nil
			if ( MiniMap_Scroll_Map_Pins:GetNamedChild( pinName ) == nil ) then
				pin = WINDOW_MANAGER:CreateControl( string.format( 'MiniMap_Scroll_Map_Pins%s', pinName ), MiniMap_Scroll_Map_Pins, CT_TEXTURE )
			else
				pin = MiniMap_Scroll_Map_Pins:GetNamedChild( pinName )
			end

			-- set defaults
			local unitX      = nil
			local unitY      = nil
			local pinWidth   = 32
			local pinHeight  = 32
			local pinScale   = false
			local pinTexture = ''
			local pinLayer   = DL_CONTROLS
			local pinColor   = { wpin:GetNamedChild( 'Background' ):GetColor() }
			pinColor         = { r = pinColor[1], g = pinColor[2], b = pinColor[3] }

			-- check pin status
			if ( wpin.m_Pin:IsQuest() ) then
				pinLayer = DL_TEXT
			end

			-- get pin texture, size, scaling
			pinWidth, pinHeight = wpin:GetDimensions()
			if ( wpin.m_Pin.m_PinType ~= nil ) then
				if ( wpin.m_Pin.pinBlobKey ~= nil ) then
					pinTexture = wpin.m_Pin.pinBlobTexture
					pinWidth,
					pinHeight  = wpin.m_Pin.pinBlob:GetDimensions()
					pinScale   = true
				elseif ( type( wpin.m_Pin.PIN_DATA[ wpin.m_Pin.m_PinType ].texture ) == 'string' ) then
					pinTexture = wpin.m_Pin.PIN_DATA[ wpin.m_Pin.m_PinType ].texture
				elseif ( type( wpin.m_Pin.PIN_DATA[ wpin.m_Pin.m_PinType ].texture ) == 'function' ) then
					pinTexture = wpin.m_Pin.PIN_DATA[ wpin.m_Pin.m_PinType ].texture( wpin.m_Pin )
				end
			end

			-- update group member pins
			local unitTag = wpin.m_Pin:GetUnitTag()
			if ( string.find( unitTag, 'group' ) ) then
				if ( pName ~= GetRawUnitName( unitTag ) and GetUnitZone( unitTag ) == pZone ) then
					unitX, unitY = GetMapPlayerPosition( unitTag )
					wpin:SetHidden( false )
					pinColor = ( IsUnitGroupLeader( unitTag ) ) and _LUI.config.colors.GROUPLEADER or _LUI.config.colors.GROUPMEMBER
					pinLayer = DL_OVERLAY
				end
			end

			-- calc pin coords
			local nX   = unitX or wpin.m_Pin.normalizedX
			local nY   = unitY or wpin.m_Pin.normalizedY
			local pinX = nX * self.zoom * self.data.map.width
			local pinY = nY * self.zoom * self.data.map.height

			-- setup display
			pin:SetHidden( not self:GetPinVisibility( wpin, pinTexture ) )
			pin:SetTexture( pinTexture )
			pin:SetColor( pinColor.r, pinColor.g, pinColor.b, 1.0 )
			pin:SetDimensions( self:GetPinSize( pinWidth, pinHeight, pinScale ) )
			pin:SetAnchor( CENTER, MiniMap_Scroll_Map_Pins, TOPLEFT, pinX, pinY )
			pin:SetDrawLayer( pinLayer )

			-- set other values
			pin.Zone = self.data.map.rawName
		end
	end
end

function __MiniMap:Zoom( delta )
	if ( delta == 0 ) then
		self.zoom = self.savedVars.defaultZoom
	else
		self.zoom = self.zoom + ( delta / 10 )
		self.zoom = ( self.zoom <= 0.1 ) and 0.1 or self.zoom
		self.zoom = ( self.zoom >= 1.8 ) and 1.8 or self.zoom
	end
	MiniMap_ZoomLabel:SetText( string.format( '%.0f%%', self.zoom * 100 ) )
end

function __MiniMap:GetPinVisibility( wpin, texture )
	return ( wpin.m_Pin.normalizedX ~= nil and wpin.m_Pin.normalizedX > 0 and texture ~= '' ) and true or false
end

function __MiniMap:GetPinSize( pinWidth, pinHeight, pinScale )
	local pinX = ( pinScale ) and self.zoom * pinWidth or pinWidth
	local pinY = ( pinScale ) and self.zoom * pinHeight or pinHeight
	return pinX * self.savedVars.defaultPinScale, pinY * self.savedVars.defaultPinScale
end

function __MiniMap:GetSettings( LAM, LAMPanelID, addOnName )
	LAM:AddSlider( LAMPanelID, string.format( '%s_SettingsMiniMapZoomSlider', addOnName ), 'Default zoom %', '', 10, 180, 10,
		function() return self.savedVars.defaultZoom * 100 end,
		function( value ) self.savedVars.defaultZoom = value / 100 end,
	false, '' )
	LAM:AddSlider( LAMPanelID, string.format( '%s_SettingsMiniMapPinScaleSlider', addOnName ), 'Default pin scale %', '', 10, 200, 10,
		function() return self.savedVars.defaultPinScale * 100 end,
		function( value ) self.savedVars.defaultPinScale = value / 100 end,
	false, '' )
end